unit UResearchUpgrade;

interface

const
  // Returned value if upgrade is not possible or wrong param values
  NO_PRICE_AVAILABLE = -1;
  // Returned value if research is not possible
  NO_RESEARCH_AVAILABLE = -1;

  // Limit for gold which can be spent for upgrade (one step)
  MAXIMUM_PRICE = 1000;

type
  TUpgradeTypes = (
    utHitpoints = 0, // 32
    utArmor,         // 32
    utAmmo,          // 32
    utAttack,        // 16
    utSpeed,         // 16
    utShots,         // 16
    utRange,         // 8
    utScan,          // 8
    utCost           // 8
  );

const
  UT_FACTORS: array [TUpgradeTypes] of integer =
    (32, 32, 32, 16, 16, 16, 8, 8, 8);

  //////////////////////////////////// Research by Laboratory //////////////////////////////////////

  // Calculates the change of the given BaseValue, with the given ResearchLevel
  // This change is independent of the UpgradeType, only utCost has a special
  // handling, because it actually decreases the value (so you will get
  // a negative value as return value).
  // Return the change of the BaseValue (can be negative if Cost is
  // the UpgradeType)
  // BaseValue the value, the unit has in it's base version
  // ResearchLevel the level for which you want to know the change
  // (e.g. 10 for 10%)
  // UpgradeType optional,set it to Cost if you need to know the changes in cost
  // MatPerTurn optional, needed for UpgradeType Cost because the behaviour
  // depend from this value. This is how much res spend unit builder
  // (1 - Barrak, 2 - Constructor or Engineer, 3 - Plant)
  function CalcChangeByResearch(
    const BaseValue, ResearchLevel: integer;
    const UpgradeType: TUpgradeTypes = utHitpoints;
    const MatPerTurn: integer = 2): integer;

  // Calculates the turns needed for one research center to reach the next level
  // Return the turns needed to reach the next level with one research center or
  // NO_RESEARCH_AVAILABLE if the passed values are out of range
  // ResearchLevel the level this research area currently has (e.g. 20 for 20%)
  // UpgradeType the area of the upgrade
  // DeltaLevel optional, how increase ResearchLevel. If 0 then calculated
  // (default is 10)
  function CalcResearchTurns(
    const ResearchLevel: Integer; const UpgradeType: TUpgradeTypes;
    const DeltaLevel: integer = 0): integer;

  // Calculates the increase of a research level, for next step.
  // Return the increase of the research level, when a research started
  function CalcIncreaseByResearch(): integer;

  // Calculate the absolute turns for achive level ResearchLevel from 0.
  // Return turns for achive level
  // ResearchLevel the level this research area currently has (e.g. 20 for 20%)
  // UpgradeType the area of the upgrade
  function CalcAbsoluteResearch(
    const ResearchLevel: Integer; const UpgradeType: TUpgradeTypes): integer;

  //////////////////////////////////// Upgrades by Gold ////////////////////////////////////////////

  // Calculates the price (gold) to upgrade from the given value.
  // Return the costs for this upgrade or NO_PRICE_AVAILABLE if no upgrade
  // CurValue the value the unit currently has (without research bonus but with
  // clan bonus)
  // BaseValue the value the unit has as a base value (with clan bonus)
  // UpgradeType the area of the upgrade
  // ResearchLevel the research level of the player for current upgrade type
  function CalcPrice(
    const CurValue, BaseValue: integer;
    const UpgradeType: TUpgradeTypes;
    const ResearchLevel: integer): integer;

  // Calculates the price (gold) for upgrading a unit, that started with
  // BaseValue and has currently CurValue, to NewValue.
  // Return the costs for this upgrade or NO_PRICE_AVAILABLE if such an upgrade
  // is impossible.
  // BaseValue the value the unit has as a base value
  // CurValue the value the unit currently has
  // NewValue the value the unit wants to reach
  // ResearchLevel the research level of the player that has to be taken into
  // account
  // UpgradeType the area of the upgrade
  function GetCostForUpgrade(
    const BaseValue, CurValue, NewValue, ResearchLevel: integer;
    const UpgradeType: TUpgradeTypes): integer;

  // Round the cost based on MatPerTurn value.
  // Return rounded value. Can not return 0.
  // Cost the value get during floating point calculation
  // MatPerTurn how much res spend unit builder per turn
  function RoundCost(
    const Cost: double; const MatPerTurn: integer = 2): integer;

  // Calculates the increase of a unit value, when an upgrade is bought.
  // Examples: If BaseValue is 10, the increase will be 2.
  // If BaseValue is 28, the increase will be 5.
  // The increase is not growing, if the unit has already some upgrades!
  // The only needed thing for the calculation is the value, at which the unit
  // started.
  // Return the increase of the unit's value, when an upgrade is bought
  // BaseValue the value, the unit has in it's base version
  function CalcIncreaseByUpgrade(const BaseValue: integer): integer;

  // Calculates the absolute cost of Value
  // Return price
  // GoldValue the value, the unit has made upgrades with gold only (exclude
  // research with Lab)
  // BaseValue the value, the unit has in it's base version
  // UpgradeType the area of the upgrade
  function CalcAbsolutePrice(
    const GoldValue, BaseValue: integer;
    const UpgradeType: TUpgradeTypes): double;

  //////////////////////////////////// Common Utilites /////////////////////////////////////////////

  // Calculates the raw-material needed for upgrading a unit, that costs
  // UnitCost, to the current version.
  // Return the raw-material needed to upgrade to the current version
  // The costs in original M.A.X. are simply a fourth of the costs (rounded
  // down) needed to build that unit.
  // The costs do not depend on the quality of the upgrade (e.g. upgrading
  // hitpoints from 18 to 20 costs the same as upgrading the basic version of
  // the unit to an ultra fat version with all values upgraded to a  maximum).
  // Return the materisal needed for upgrade a unit
  // UnitCost the raw-material cost to build the unit that will be upgraded
  // (e.g. 24 for a Mine)
  function GetMaterialCostForUpgrading(const UnitCost: integer): integer;


implementation

uses
  Math;

function CalcPrice(
  const CurValue, BaseValue: integer;
  const UpgradeType: TUpgradeTypes;
  const ResearchLevel: integer): integer;
var
  bonusByResearch: integer;
  goldOnlyValue: integer;
  price: integer;
  dValue: integer;
begin
  bonusByResearch :=
    CalcChangeByResearch(BaseValue, ResearchLevel);
  goldOnlyValue := CurValue - bonusByResearch;

  dValue := CalcIncreaseByUpgrade(BaseValue);

  price :=
    Trunc(
      CalcAbsolutePrice(goldOnlyValue + dValue, BaseValue, UpgradeType) -
      CalcAbsolutePrice(goldOnlyValue, BaseValue, UpgradeType));

  if price < MAXIMUM_PRICE then
    Result := price
  else
    Result := NO_PRICE_AVAILABLE;
end;

function GetCostForUpgrade(
  const BaseValue, CurValue, NewValue, ResearchLevel: integer;
  const UpgradeType: TUpgradeTypes): integer;
var
  cost, dCost: integer;
  upgradedValue: integer;
begin
  cost := 0;
  if (BaseValue <= CurValue) and (CurValue < NewValue) then begin
    upgradedValue := CurValue;
    while upgradedValue < NewValue do begin
      dCost := CalcPrice(upgradedValue, BaseValue, UpgradeType, ResearchLevel);
      if dCost <> NO_PRICE_AVAILABLE then begin
        cost := cost + dCost;
        upgradedValue := upgradedValue + CalcIncreaseByUpgrade(BaseValue);
        if upgradedValue > NewValue then begin
          cost := NO_PRICE_AVAILABLE;
          Break;
        end;
      end
      else begin
        cost := NO_PRICE_AVAILABLE;
        Break;
      end;
    end;
  end;
  Result := cost;
end;

function CalcChangeByResearch(
  const BaseValue, ResearchLevel: integer;
  const UpgradeType: TUpgradeTypes = utHitpoints;
  const MatPerTurn: integer = 2): integer;
var
  cost: double;
begin
  if ResearchLevel <=0 then begin // no research done yet...
    Result := 0;
    Exit;
  end;
  if UpgradeType <> utCost then begin // standard research areas - all handled the same way
    Result := Trunc((BaseValue * (100 + ResearchLevel)) / 100) - BaseValue;
  end
  else begin
    cost := BaseValue * 100 / (100 + ResearchLevel);
    // now the real cost is rounded to the next possible cost value
    // (Unit factories: steps of 3, Building construction: steps of 2,
    // Infantry training: steps of 1)
    Result := RoundCost(cost, MatPerTurn) - BaseValue;
  end;
end;

function CalcAbsoluteResearch(
  const ResearchLevel: Integer; const UpgradeType: TUpgradeTypes): integer;
begin
  Result := Trunc(Power(1 + ResearchLevel / 100, 7.5) * 256 / UT_FACTORS[UpgradeType]);
end;

function CalcIncreaseByResearch(): integer;
begin
  Result := 10;
end;

function CalcResearchTurns(
  const ResearchLevel: integer; const UpgradeType: TUpgradeTypes;
  const DeltaLevel: integer = 0): integer;
var
  dLevel: integer;
begin
  if DeltaLevel = 0 then
    dLevel := CalcIncreaseByResearch()
  else
    dLevel := DeltaLevel;
  if (DeltaLevel < 0) or (ResearchLevel < 0) then
    Result := NO_RESEARCH_AVAILABLE
  else
    Result :=
      CalcAbsoluteResearch(ResearchLevel + dLevel, UpgradeType) -
      CalcAbsoluteResearch(ResearchLevel,          UpgradeType);
end;

function RoundCost(const Cost: double; const MatPerTurn: integer = 2): integer;
begin
  if MatPerTurn <= 0 then
    Result := Trunc(Cost)
  else
    Result := Round(Cost / MatPerTurn) * MatPerTurn;

  if Result <= 0 then // a cost of zero or below is forbidden
    Result := MatPerTurn;
end;

function CalcIncreaseByUpgrade(const BaseValue: integer): integer;
begin
  if BaseValue < 10 then
    Result := 1
  else if BaseValue < 26 then
    Result := 2
  else if BaseValue < 56 then
    Result := 5
  else
    Result := 10;
end;

function CalcAbsolutePrice(
  const GoldValue, BaseValue: integer;
  const UpgradeType: TUpgradeTypes): double;
begin
  Result := Power(GoldValue / BaseValue, 7.5) * 64 / UT_FACTORS[UpgradeType];
end;

function GetMaterialCostForUpgrading(const UnitCost: integer): integer;
begin
  if UnitCost < 4 then
    Result := 0
  else
    Result := UnitCost div 4;
end;

end.
